@@ -1,6 +1,6 @@
 ##############################################################################
 #
-# Copyright (c) 2008 Zope Corporation and Contributors.
+# Copyright (c) 2008 Zope Foundation and Contributors.
 # All Rights Reserved.
 #
 # This software is subject to the provisions of the Zope Public License,
@@ -15,9 +15,28 @@
 import transaction as zope_transaction
 from zope.interface import implements
 from transaction.interfaces import ISavepointDataManager, IDataManagerSavepoint
+from transaction._transaction import Status as ZopeStatus
+from sqlalchemy.orm.exc import ConcurrentModificationError
+from sqlalchemy.exc import DBAPIError
 from sqlalchemy.orm.session import SessionExtension
 from sqlalchemy.engine.base import Engine
 
+_retryable_errors = []
+try:
+    import psycopg2.extensions
+except ImportError:
+    pass
+else:
+    _retryable_errors.append((psycopg2.extensions.TransactionRollbackError, None))
+
+# ORA-08177: can't serialize access for this transaction
+try:
+    import cx_Oracle
+except ImportError:
+    pass
+else:
+    _retryable_errors.append((cx_Oracle.DatabaseError, lambda e: e.args[0].code==8177))
+
 # The status of the session is stored on the connection info
 STATUS_ACTIVE = 'active' # session joined to transaction, writes allowed.
 STATUS_CHANGED = 'changed' # data has been written
@@ -43,7 +62,8 @@ class SessionDataManager(object):
     
     implements(ISavepointDataManager)
 
-    def __init__(self, session, status):
+    def __init__(self, session, status, transaction_manager):
+        self.transaction_manager = transaction_manager
         self.tx = session.transaction._iterate_parents()[-1]
         self.session = session
         _SESSION_STATE[id(session)] = status
@@ -108,6 +128,18 @@ class SessionDataManager(object):
     
     def _savepoint(self):
         return SessionSavepoint(self.session)
+    
+    def should_retry(self, error):
+        if isinstance(error, ConcurrentModificationError):
+            return True
+        if isinstance(error, DBAPIError):
+            orig = error.orig
+            for error_type, test in _retryable_errors:
+                if isinstance(orig, error_type):
+                    if test is None:
+                        return True
+                    if test(orig):
+                        return True
 
 
 class TwoPhaseSessionDataManager(SessionDataManager):
@@ -145,7 +177,7 @@ class SessionSavepoint:
         self.transaction.rollback()
 
 
-def join_transaction(session, initial_state=STATUS_ACTIVE):
+def join_transaction(session, initial_state=STATUS_ACTIVE, transaction_manager=zope_transaction.manager):
     """Join a session to a transaction using the appropriate datamanager.
        
     It is safe to call this multiple times, if the session is already joined
@@ -154,20 +186,25 @@ def join_transaction(session, initial_state=STATUS_ACTIVE):
     `initial_state` is either STATUS_ACTIVE, STATUS_INVALIDATED or STATUS_READONLY
     
     If using the default initial status of STATUS_ACTIVE, you must ensure that
-    dirty_session(session) is called when data is written to the database.
+    mark_changed(session) is called when data is written to the database.
     
-    The DirtyAfterFlush SessionExtension can be used to ensure that this is
+    The ZopeTransactionExtesion SessionExtension can be used to ensure that this is
     called automatically after session write operations.
     """
     if _SESSION_STATE.get(id(session), None) is None:
-        DataManager = session.twophase and TwoPhaseSessionDataManager or SessionDataManager
-        zope_transaction.get().join(DataManager(session, initial_state))
+        if session.twophase:
+            DataManager = TwoPhaseSessionDataManager
+        else:
+            DataManager = SessionDataManager
+        transaction_manager.get().join(DataManager(session, initial_state, transaction_manager))
 
-def mark_changed(session):
-    """Mark a session as needing to be committed
+def mark_changed(session, transaction_manager=zope_transaction.manager):
+    """Mark a session as needing to be committed.
     """
-    assert _SESSION_STATE[id(session)] is not STATUS_READONLY
-    _SESSION_STATE[id(session)] = STATUS_CHANGED
+    session_id = id(session)
+    assert _SESSION_STATE.get(session_id, None) is not STATUS_READONLY, "Session already registered as read only"
+    join_transaction(session, STATUS_CHANGED, transaction_manager)
+    _SESSION_STATE[session_id] = STATUS_CHANGED
 
 
 class ZopeTransactionExtension(SessionExtension):
@@ -175,25 +212,26 @@ class ZopeTransactionExtension(SessionExtension):
     the DataManager to rollback rather than commit on read only transactions.
     """
     
-    def __init__(self, initial_state=STATUS_ACTIVE):
+    def __init__(self, initial_state=STATUS_ACTIVE, transaction_manager=zope_transaction.manager):
         if initial_state=='invalidated': initial_state = STATUS_CHANGED #BBB
         SessionExtension.__init__(self)
         self.initial_state = initial_state
+        self.transaction_manager = transaction_manager
     
     def after_begin(self, session, transaction, connection):
-        join_transaction(session, self.initial_state)
+        join_transaction(session, self.initial_state, self.transaction_manager)
     
     def after_attach(self, session, instance):
-        join_transaction(session, self.initial_state)
+        join_transaction(session, self.initial_state, self.transaction_manager)
     
     def after_flush(self, session, flush_context):
-        mark_changed(session)
+        mark_changed(session, self.transaction_manager)
         
     def after_bulk_update(self, session, query, query_context, result):
-        mark_changed(session)
+        mark_changed(session, self.transaction_manager)
 
     def after_bulk_delete(self, session, query, query_context, result):
-        mark_changed(session)
+        mark_changed(session, self.transaction_manager)
     
     def before_commit(self, session):
-        assert zope_transaction.get().status == 'Committing', "Transaction must be committed using the transaction manager"
+        assert self.transaction_manager.get().status == ZopeStatus.COMMITTING, "Transaction must be committed using the transaction manager"
